/*
	hier.c - Module to implement hierarchical structuring into rtftohtml.

	Author: 	Christian Bolik (zzhibol@rrzn-user.uni-hannover.de)

	Started: 	05.05.94
	Last changed:	25.07.94
	Version:	1.2

	ChangeLog:
			06.07.94 updated to work with rtftohtml-2.7 --> V1.1
			21.07.94 no more ':' in filenames (newHierFile),
				 new title for html-documents
			25.07.94 improved -h0 behaviour, bug fixes --> V1.2

	Functions:
			hierCreationString ()
			doHierPass ()
			hierControl ()
			hierText ()
			hierGroup ()
			hierHeadingLevel(style_num)
			newHeading ()
			hierProcess ()
			newHierFile (heading)
			fsuffix (filenum)
			hierOutFile ()
			hierProlog (outfd)
			hierExtList (hh, outfd)
			hierIntList (hh, outfd)
			hierEpilog (outfd)
			hierChangeOutFile ()
			generateContents ()
			hierContList (hh, outfd)
			setNavTexts (lang)
			hierNavPanel (outfd)
			colorIsRed (colnum)
			hierResolve ()
			addIndexEntry ()
			newIndexEntry ()
			generateIndex ()
			sortIndex ()
			alpha_strcmp (s1, s2)
*/

#ifdef CB_ADD

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
# define Edef
#include "rtf.h"
#include "rtftohtml.h"
#include "hier.h"

char rtftowebVersion[6] = "1.2";

int hierPass = 0;
int inHeading = 0;
int inHeadingRef = 0;
int groupInHeading = 0;
int inIndexEntry = 0;
int hierDepth = 9999;	/* this means: split to maximal depth */
int hierHasText = 0;
int hierContents = 0;
int hierRefsOnTop = 0;
int hierIndex = 0;
int hierIsChangingFile = 0;

char headingRef[80];
char indexEntry[80];

FILE *hierFile = NULL;

static HierHeading *hierList = NULL;
static HierFile *hierFileList = NULL;
static HierIndex *hierIndexList = NULL;

static void hierControl (/*void*/);
static void hierText (/*void*/);
static void hierGroup (/*void*/);
static HierHeading *newHeading (/*void*/);
static void hierProcess (/*void*/);
static HierFile *newHierFile (/*HierHeading *heading*/);
static char *fsuffix (/*int filenum*/);

static void hierExtList (/*HierHeading *hh, FILE *outfd*/);
static HierHeading *hierIntList (/*HierHeading *hh, FILE *outfd*/);
static void generateContents (/*void*/);
static HierHeading *hierContList (/*HierHeading *hh, FILE *outfd*/);

static void setNavTexts (/*int lang*/);
static int hierNavPanel (/*FILE *outfd*/);

static HierIndex *newIndexEntry (/*void*/);
static HierIndex *sortIndex (/*void*/);
static int alpha_strcmp (/*char *s1, char *s2*/);

static int inTitle = 0;
static int indexInHeading = 0;
static hiddenText = 0;
static HierHeading *currentHH = NULL;
static char docTitle[128] = "";
static char navText[NAV_NUMBER][80];

static int num_headings = 0;
static int num_files = 0;
static int num_index_entries = 0;

static int file_number = -1;	/* number of opened files - 1 */
static int heading_number = 0;	/* number of read headings */

/* ========================================================================== */

char *hierCreationString ()
{
static char str[160];

   sprintf (str, "<!-- This document was created from RTF source by rtftohtml version %s,\n     extended by rtftoweb version %s. -->\n", 
		PVers, rtftowebVersion);

   return (str);
} /* hierCreationString */

void doHierPass ()
{
   /* the RTF-file has already been opened */
   /* Now prepare for the preprocessing pass: */
   
   /* install class callbacks */
   RTFSetClassCallback (rtfControl, hierControl);
   RTFSetClassCallback (rtfText, hierText);
   RTFSetClassCallback (rtfGroup, hierGroup);

   /* de-install default destination callbacks */
   RTFSetDestinationCallback (rtfInfo, NULL);
   RTFSetDestinationCallback (rtfPict, NULL);
   RTFSetDestinationCallback (rtfObject, NULL);

   setNavTexts (DEFAULT_LANG);

   /* now preprocess the input */
   RTFRead ();
   if (!strlen(docTitle)) {
      strcat (docTitle, "Without a title");
   }
#  ifdef DEBUG
   fprintf(stderr, "docTitle: %s\n", docTitle);
#  endif

   /* fill required data structures */
   hierProcess();

   /* clean things up */
   inHeading = 0;
   rewind (stdin);
   RTFInit ();
} /* doHierPass */

/* -------------------------------------------------------------------------- */

static void hierControl ()
{
int level;

   switch (rtfMajor) {

   case rtfParAttr:
      /* Paragraph Style Token */
      switch (rtfMinor) {
      case rtfParDef:
	 if (inHeading && !groupInHeading) {
	    inHeading = 0;
#           ifdef DEBUGW
	    fprintf(stderr, "%s\n", currentHH->text);
	    fprintf(stderr, "Length: %d\n", strlen (currentHH->text));
	    fprintf(stderr, "rtfParDef encountered\n");
#           endif
	 }
	 break;
      case rtfStyleNum:
	 if ((level = hierHeadingLevel (rtfParam)) && !inHeading) {
	    inHeading = 1;
	    currentHH = newHeading ();
	    currentHH->level = level;
	 }
#        ifdef DEBUG
	 fprintf(stderr, "rtfStyleNum: %d = %s, Level %d\n", rtfParam, 
		RTFGetStyle(rtfParam)->rtfSName, currentHH->level);
#        endif
	 break;
      }
      break;

   case rtfCharAttr:
      /* look for language of RTF-file */
      switch (rtfMinor) {

      case rtfInvisible:
         hiddenText = 1;
	 break;
      case rtfLanguage:
	 setNavTexts (rtfParam);
	 break;

      }
      break;

   case rtfSpecialChar:
      /* convert tabs in headings to spaces */
      if (inHeading)
	 if (rtfMinor == rtfTab)
	    strcat (currentHH->text, " ");
      break;

   case rtfDestination:
      switch (rtfMinor) {

      case rtfIndex:
	 if (hierIndex && inHeading) 
	    /* catch index entries that appear in headings */
	    indexInHeading = 1;
	 break;
      case rtfITitle:
	 inTitle = 1;
#        ifdef DEBUG
	 fprintf(stderr, "inTitle = 1\n");
#        endif
	 break;
      default:
	 inTitle = 0;

      }
      break;
   }
} /* hierControl */

/* -------------------------------------------------------------------------- */

static void hierText ()
{
   if (inHeading) {
      if (!hiddenText) 
         strcat (currentHH->text, MapChar (rtfMinor));
      if (indexInHeading)
	 strcat (indexEntry, MapChar (rtfMinor));
   } else if (inTitle) {
      strcat (docTitle, MapChar (rtfMinor));
   }

} /* hierText */

/* -------------------------------------------------------------------------- */

static void hierGroup ()
{
static int groupInHeadingCount = 0, 
	   groupInIndexCount = 0,
	   hiddenTextCount = 0;

   switch (rtfMajor) {
      case rtfBeginGroup:
	 if (inHeading) {
	    groupInHeading = 1;
	    groupInHeadingCount++;
	 }
	 if (indexInHeading)
	    groupInIndexCount++;
	 if (hiddenText)
	    hiddenTextCount++;
	 break;
      case rtfEndGroup:
	 if (groupInHeadingCount) {
	    groupInHeadingCount--;
	 }
	 if (groupInHeadingCount == 0 && groupInHeading)
	    groupInHeading = 0;
	 if (indexInHeading) {
	    if (groupInIndexCount)
	       groupInIndexCount--;
	    else {
	       indexInHeading = 0;
	       addIndexEntry()->heading = currentHH;
	       currentHH->index = 1;
	    }
	 }
	 if (hiddenText) {
	    if (hiddenTextCount)
	       hiddenTextCount--;
	    else
	       hiddenText = 0;
	 }
	 break;
   }
} /* hierGroup */

/* -------------------------------------------------------------------------- */

int hierHeadingLevel(style_num)
   int style_num;
/* Returns the level of style style_num if it's a heading, otherwise 0 */
{
char style_name[64];
int level;

   strcpy(style_name, RTFGetStyle(style_num)->rtfSName);

   /* Check for heading style; of course this should be generalized
      using html-trans !! */
   if (!strncmp ("heading", style_name, 7)) {
      level = atoi (&style_name[8]);
   } else
      level = 0;

   return (level);
} /* hierHeadingLevel */

/* -------------------------------------------------------------------------- */

static HierHeading *newHeading ()
{
HierHeading *newhh, *hh;

   newhh = (HierHeading *) RTFAlloc (sizeof (HierHeading));
#ifdef DEBUGM
   fprintf (stderr, "RTFAlloc in newHeading returns: %p\n", newhh);
#endif
   if (!newhh)
      RTFPanic("Not enough memory in newHeading!\n");

   newhh->next = newhh->peer = newhh->father = newhh->fchild = NULL;
   newhh->index = 0;
   newhh->level = 1;
   strcpy(newhh->text, "");

   hh = hierList;
   if (hh == NULL)
      hierList = newhh;
   else {
      while (hh->next)
	 hh = hh->next;
      hh->next = newhh;
   }

   return (newhh);
} /* HierHeading */

/* -------------------------------------------------------------------------- */

static void hierProcess ()
/* processes the filtered hierarchy */
{
HierHeading *hh = hierList, *ihh, *jhh;
HierIndex *hi;

   while (hh) {
      ihh = hh->next;

      /* find fchild and peers */

      while (ihh) {
	 if (!hh->fchild && ihh->level > hh->level) {
	    hh->fchild = ihh;
            ihh->father = hh;
	    jhh = ihh->next;
	    while (jhh) {
		if (jhh->level == ihh->level)
		   	jhh->father = ihh->father;
                if (jhh->level < ihh->level)
			break;
		jhh = jhh->next;
	    }
         }
	 if (ihh->level == hh->level) {
	    hh->peer = ihh;
	    break;
	 }
	 if (ihh->level < hh->level)
	    break;

	 ihh = ihh->next;
      }

      num_headings++;
      hh->number = num_headings;
      hh = hh->next;	
   }

   num_files = 1;
   hh = hierList;
   while (hh) {
      if (hh->level <= hierDepth) {
	 newHierFile (hh);
	 num_files++;
      }
      hh->filenum = num_files - 1;
      hh = hh->next;
   }
#  ifdef DEBUG
   fprintf (stderr, "num_headings: %d, num_files: %d\n", 
		num_headings, num_files);
#  endif

   if (hierIndexList) {
      /* oh, there have been indexes in headings, postprocess them */
      hi = hierIndexList;
      while (hi) {
	 if (hi->heading) {
	    hi->filenum = hi->heading->filenum;
	 }
	 hi = hi->next;
      }
   }

} /* hierProcess */

/* -------------------------------------------------------------------------- */

static HierFile *newHierFile (heading)
   HierHeading *heading;
{
HierFile *newhf, *hf;
char word1[FSUFFIXSIZE + 4], *fs;
int used;

   newhf = (HierFile *) RTFAlloc (sizeof (HierFile));
#ifdef DEBUGM
   fprintf (stderr, "RTFAlloc in newHierFile returns: %p\n", newhf);
#endif
   if (!newhf)
      RTFPanic("Not enough memory in newHierFile!\n");

   newhf->next = NULL;
   newhf->heading = heading;

   /* find a descriptive suffix for this file's name */
   word1[FSUFFIXSIZE] = 0;
   strncpy (word1, heading->text, FSUFFIXSIZE);
   if (strlen (word1) == FSUFFIXSIZE) {
      /* find first space and cut it off */
      fs = strchr (word1, ' ');
      if (fs)
	 *fs = 0;
   } /* else {
      fs = word1;
      while (*fs) {
	 if (*fs == ' ')
	    *fs = '_';
	 fs++;
      }
   } */

   /* replace critical characters to underlines */
   fs = word1;
   while (*fs) {
      if (*fs == ' ')
         *fs = '_';
      else if (*fs == ':')	/* Mosaic doesn't like colons in filenames */
         *fs = '_';
      fs++;
   }
   sprintf (newhf->fsuffix, "-%s", word1);

   /* now see if this suffix is already used */
   if (!strcmp (newhf->fsuffix, "-Contents"))
      strcat (newhf->fsuffix, "-2");
   if (!strcmp (newhf->fsuffix, "-Index"))
      strcat (newhf->fsuffix, "-2");
   used = 1;
   hf = hierFileList;
   while (hf) {
      while (hf) {
         if (!strcmp (newhf->fsuffix, hf->fsuffix)) {
	    used++;
	    sprintf (newhf->fsuffix, "%s-%d", word1, used);
	    hf = hierFileList;
	    break;
         }
         hf = hf->next;
      }
   }

   hf = hierFileList;
   if (hf == NULL) {
      hierFileList = newhf;
   } else {
      while (hf->next)
	 hf = hf->next;
      hf->next = newhf;
   }

   return (newhf);
}

/* -------------------------------------------------------------------------- */

static char *fsuffix (filenum)
   int filenum;
/* returns fsuffix for a given file-number */
{
HierFile *hf = hierFileList;
int i;

   if (filenum == 0)
      return ("");

   for (i = 1; i < filenum; i++)
      hf = hf->next;

   return (hf->fsuffix);
} /* fsuffix */

/* ========================================================================== */

char *hierOutFile ()
{
HierFile *hf = hierFileList;
char fname[128];
int i;

   file_number++;

   if (file_number > 0) {
      for (i = 1; i < file_number; i++)
	 hf = hf->next;
      sprintf(fname, "%s%s.html", FPrefix, hf->fsuffix);
   } else
      sprintf(fname, "%s.html", FPrefix);

#  ifdef DEBUG
   fprintf (stderr, "hierOutFile: %s\n", fname);
#  endif

   return (fname);
} /* hierOutFile */

/* -------------------------------------------------------------------------- */

void hierProlog (outfd)
   FILE *outfd;
/* Write a prolog to file outfd with number file_number. */
{
HierHeading *hh;
HierFile *hf;
int i;

   fprintf (outfd, "<html><head>\n");
   fprintf (outfd, "%s", hierCreationString ());
    if (file_number == 0) 
	fprintf (outfd, "<title>%s - Title</title></head><body>\n", docTitle);
    else {
	hf = hierFileList;
	for (i = 1; i < file_number; i++)
	    hf = hf->next;
	fprintf (outfd, "<title>%s - %s</title></head><body>\n", 
		docTitle, hf->heading->text);
    }

   if (hierNavPanel (outfd))
      fprintf (outfd, "<hr>\n");

   if (file_number == 0) {
      /* Aha, this is the toplevel file (title page) */
      fprintf (outfd, "<h1>\n%s</h1>\n", docTitle);

      if ((hh = hierList)) {
	 if (hierDepth == 0)
	    hierIntList (hh, outfd);
	 else if (hierRefsOnTop)
	    hierExtList (hh, outfd);

	 if (hierDepth == 0 || hierRefsOnTop) {
	    if (hierContents) {
	       generateContents ();
      	       fprintf (outfd, 
                   "<ul>\n<li><a href=\"%s-Contents.html\"> %s </a>\n</ul>\n",
 			FPrefix, navText[NAV_CONTENTS]);
	    }

	    fprintf (outfd, "<hr>\n");
         }
	 hierHasText = 0;
      } else
	 hierHasText = 1;
   } else {
      fprintf (outfd, "<h%d>\n%s</h%d>\n", 
		hf->heading->level, hf->heading->text, hf->heading->level);

      if ((hh = hf->heading->fchild)) {
	 if (hierDepth < hh->level)
	    hierIntList (hh, outfd);
	 else if (hierRefsOnTop)
	    hierExtList (hh, outfd);

	 if (hierDepth < hh->level || hierRefsOnTop)
	    fprintf (outfd, "<hr>\n");

	 hierHasText = 0;
      } else
	 hierHasText = 1;
   }
} /* hierProlog */

/* -------------------------------------------------------------------------- */

static void hierExtList (hh, outfd)
   HierHeading *hh;
   FILE *outfd;
/* write list of external links to headings */
{
   fprintf (outfd, "<ul>\n");      
   while (hh) {
      fprintf (outfd, "<li><a href=\"%s%s.html\"> %s </a>\n", 
			FPrefix, fsuffix (hh->filenum), hh->text);
      hh = hh->peer;
   }
   fprintf (outfd, "</ul>\n");      
} /* hierExtList */

/* -------------------------------------------------------------------------- */

static HierHeading *hierIntList (hh, outfd)
   HierHeading *hh;
   FILE *outfd;
/* write list of internal links to headings */
{
int level = hh->level; 

   fprintf (outfd, "<ul>\n");      
   while (hh) {
      fprintf (outfd, "<li><a href=\"#Heading%d\"> %s </a>\n", 
			hh->number, hh->text);
      hh = hh->next;
      if (hh) {
         if (hh->level > level)
	    hh = hierIntList (hh, outfd);
         if (hh->level < level)
	    break;
      }
   }
   fprintf (outfd, "</ul>\n");      
   return (hh);
} /* hierIntList */

/* -------------------------------------------------------------------------- */

void hierEpilog (outfd)
   FILE *outfd;
/* Write an epilog to file outfd with number file_number. */
{
HierHeading *hh;
HierFile *hf;
int i;

   if (hierHasText && (hierDepth > 0 || hierContents || hierIndex ))
      fprintf (outfd, "<p><hr>\n");

   if (!hierRefsOnTop) {

   if (file_number == 0) {
      /* Aha, this is the toplevel file (title page) */
      if ((hh = hierList)) {
	 if (hierDepth > 0) {
	    hierExtList (hh, outfd);

	    if (hierContents) {
	       generateContents ();
      	       fprintf (outfd, 
                   "<ul>\n<li><a href=\"%s-Contents.html\"> %s </a>\n</ul>\n",
 			FPrefix, navText[NAV_CONTENTS]);
	    }

	    fprintf (outfd, "<hr>\n");
	 }
      }
   } else {
      hf = hierFileList;
      for (i = 1; i < file_number; i++)
	 hf = hf->next;

      if ((hh = hf->heading->fchild)) {
	 if (hierDepth >= hh->level) {
	    hierExtList (hh, outfd);

 	    fprintf (outfd, "<hr>\n");
	 }
      }
   }

   } /* !hierRefsOnTop */

   hierNavPanel (outfd);

} /* hierEpilog */

/* -------------------------------------------------------------------------- */

void hierChangeOutFile ()
/* RTFRead has found a heading, change output file if necessary */
{
HierHeading *hh;
int i;


   heading_number++;
   hh = hierList;
   for (i = 1; i < heading_number; i++)
      hh = hh->next;
   /* hh points to heading that has just been read */
   
   if (hh->level <= hierDepth) {
      hierIsChangingFile = 1;
      HTMLCleanup ();
      HTMLInit ();
      hierIsChangingFile = 0;
   } else {
      FlushHTML ();

      fprintf (hierFile, "\n<a name=\"Heading%d\">\n", hh->number);
      fprintf (hierFile, "<h%d> %s </h%d></a>\n", 
			hh->level, hh->text, hh->level);
   }
} /* hierChangeOutFile */

/* ========================================================================== */

static void generateContents ()
/* creates a file for the table of contents */
{
FILE *fd;
char fname[256];
int ofn;

   sprintf (fname, "%s-Contents.html", FPrefix);

   fd = fopen(fname, "w");
   if (fd == NULL) {
      RTFPanic ("Open of %s Failed", fname);
   }

   ofn = file_number;
   file_number = -1;
#ifdef DEBUG
   fprintf (stderr, "Generating Table of Contents...\n");
#endif

   fprintf (fd, "<html><head>\n");
   fprintf (fd, "%s", hierCreationString ());
   fprintf (fd, "<title> %s - %s </title></head><body>\n", docTitle, navText[NAV_CONTENTS]);

   if (hierNavPanel (fd))
      fprintf (fd, "<hr>\n");

   fprintf (fd, "<h1> %s - %s </h1>", docTitle, navText[NAV_CONTENTS]);
   hierContList (hierList, fd);

   fprintf (fd, "<hr>\n");
   hierNavPanel (fd);

   fprintf (fd, "</body></html>\n");
   fclose (fd);
   file_number = ofn;
} /* generateContents */

/* -------------------------------------------------------------------------- */

static HierHeading *hierContList (hh, outfd)
   HierHeading *hh;
   FILE *outfd;
/* write list of external links to headings for the toc */
{
int level = hh->level; 

   fprintf (outfd, "<ul>\n");      
   while (hh) {
      if (hh->level <= hierDepth)
         fprintf (outfd, "<li><a href=\"%s%s.html\"> %s </a>\n", 
			FPrefix, fsuffix (hh->filenum), hh->text);
      else
         fprintf (outfd, "<li><a href=\"%s%s.html#Heading%d\"> %s </a>\n", 
			FPrefix, fsuffix (hh->filenum), hh->number, hh->text);

      hh = hh->next;
      if (hh) {
         if (hh->level > level)
	    hh = hierContList (hh, outfd);
	 if (hh)
            if (hh->level < level)
	       break;
      }
   }
   fprintf (outfd, "</ul>\n");      
   return (hh);
} /* hierIntList */

/* ========================================================================== */

static void setNavTexts (lang)
   int lang;
{
static int oldLang = -1;

   if (lang != oldLang) {

   switch (lang) {
      case rtfLangGerman:
	 strcpy (navText[NAV_PREVIOUS], GER_PREVIOUS_TEXT);
	 strcpy (navText[NAV_NEXT], GER_NEXT_TEXT);
	 strcpy (navText[NAV_UP], GER_UP_TEXT);
	 strcpy (navText[NAV_TITLE], GER_TITLE_TEXT);
	 strcpy (navText[NAV_CONTENTS], GER_CONTENTS_TEXT);
	 strcpy (navText[NAV_INDEX], GER_INDEX_TEXT);
	 oldLang = lang;
	 break;

      default:
	 strcpy (navText[NAV_PREVIOUS], PREVIOUS_TEXT);
	 strcpy (navText[NAV_NEXT], NEXT_TEXT);
	 strcpy (navText[NAV_UP], UP_TEXT);
	 strcpy (navText[NAV_TITLE], TITLE_TEXT);
	 strcpy (navText[NAV_CONTENTS], CONTENTS_TEXT);
	 strcpy (navText[NAV_INDEX], INDEX_TEXT);
	 oldLang = lang;
   }

   }
} /* setNavTexts */

/* -------------------------------------------------------------------------- */

static int hierNavPanel (outfd)
   FILE *outfd;
{
HierFile *hf;
HierHeading *hh;
int i;
long pos;
static char navpanel[128];

   pos = ftell (outfd);

   /*fprintf (outfd, "<i>");*/ /* italic does not look too good */

   if (file_number == -2) {
      /* Index */
      if (num_files > 1) {
         fprintf (outfd, "<a href=\"%s.html\">%s</a>\n", 
			FPrefix, navText[NAV_TITLE]);
         if (hierContents)
            fprintf (outfd, ", ");
      }
   } else if (file_number == -1) {
      /* Table of Contents */
      if (num_files > 1) {
         fprintf (outfd, "<a href=\"%s.html\">%s</a>\n", 
			FPrefix, navText[NAV_TITLE]);
      }
   } else if (file_number == 0) {
      /* Title Page */
      if (num_files > 1) {
         fprintf (outfd, "<a href=\"%s%s.html\">%s</a>\n", 
			FPrefix, fsuffix (1), navText[NAV_NEXT]);
         if (hierContents)
            fprintf (outfd, ", ");
      }
   } else {
      hf = hierFileList;
      for (i = 1; i < file_number; i++)
	 hf = hf->next;

      hh = hf->heading;
      if (hh->filenum == 1)
         fprintf (outfd, "<a href=\"%s%s.html\">%s</a>\n", 
		FPrefix, fsuffix (0), navText[NAV_PREVIOUS]);
      else
         fprintf (outfd, "<a href=\"%s%s.html\">%s</a>\n", 
		FPrefix, fsuffix (hh->filenum - 1), navText[NAV_PREVIOUS]);
      if (hh->filenum < num_files - 1) {
         fprintf (outfd, ", <a href=\"%s%s.html\">%s</a>\n", 
		FPrefix, fsuffix (hh->filenum + 1), navText[NAV_NEXT]);
      } 
      if (hierDepth > 1) {
         if (hh->father)
            fprintf (outfd, ", <a href=\"%s%s.html\">%s</a>\n", 
		FPrefix, fsuffix (hh->father->filenum), navText[NAV_UP]);
         else
            fprintf (outfd, ", <a href=\"%s%s.html\">%s</a>\n", 
			FPrefix, fsuffix (0), navText[NAV_UP]);
      }
      fprintf (outfd, ", <a href=\"%s%s.html\">%s</a>\n", 
				FPrefix, fsuffix (0), navText[NAV_TITLE]);
      if (hierContents)
         fprintf (outfd, ", ");
   }

   if (hierContents && file_number != -1)
      fprintf (outfd, "<a href=\"%s-Contents.html\">%s</a>\n", 
			FPrefix, navText[NAV_CONTENTS]);

   if (hierIndex && file_number != -2) {
      if (ftell (outfd) != pos)
         fprintf (outfd, ", <a href=\"%s-Index.html\">%s</a>\n", 
			FPrefix, navText[NAV_INDEX]);
      else
         fprintf (outfd, "<a href=\"%s-Index.html\">%s</a>\n", 
			FPrefix, navText[NAV_INDEX]);
   }


   /*fprintf (outfd, "</i>\n");*/

   if (ftell (outfd) != pos)
      return (1);
   else
      return (0);

} /* hierNavPanel */

/* ========================================================================== */

int colorIsRed (colnum)
   int colnum;
{
RTFColor *colptr;

   if ((colptr = RTFGetColor (colnum))) {
      if (colptr->rtfCRed == 255 && !colptr->rtfCGreen && !colptr->rtfCGreen)
	 return(1);
      else
	 return(0);
   } else
      return(0);
} /* colorIsRed */

/* -------------------------------------------------------------------------- */

void hierResolve ()
{
HierHeading *hh = hierList, *refhh = NULL;

#ifdef DEBUG2
   fprintf (stderr, "Resolving: %s\n", headingRef);
#endif

   while (hh) {
      if (!strncmp (hh->text, headingRef, strlen (headingRef))) {
	 refhh = hh;
	 break;
      }

      hh = hh->next;
   }

   FlushHTML ();

   if (refhh) {
      if (refhh->level <= hierDepth)
         fprintf (hierFile, "<a href=\"%s%s.html\">%s</a>", 
		FPrefix, fsuffix (refhh->filenum), headingRef);
      else
         fprintf (hierFile, "<a href=\"%s%s.html#Heading%d\">%s</a>", 
		FPrefix, fsuffix (refhh->filenum), refhh->number, headingRef);
   } else {
      fprintf (stderr, "Warning: Couldn't resolve \"%s\"!\n", headingRef);
      fprintf (hierFile, "%s", headingRef);
   }
} /* hierResolve */

/* ========================================================================== */

HierIndex *addIndexEntry ()
/* adds indexEntry to list of index entries and inserts an anchor */
/* Note: This does not happen in the pre-pass, but in the main-pass! */
/* Well, for indexes in headings it actually _does_ happen int the pre-pass. */
{
HierIndex *hi;

#ifdef DEBUG2
   fprintf (stderr, "indexEntry: %s\n", indexEntry);
#endif

   hi = newIndexEntry ();
   hi->number = ++num_index_entries;
   hi->filenum = file_number;
   strcpy (hi->text, indexEntry);
   strcpy (indexEntry, "");

   /* now insert an anchor in the current html-file */
   if (hierFile) {
      FlushHTML ();
      /* Unfortunately, NCSA Mosaic depends on non-blank text between the
	 start and end tags of named anchors. By the way, Lynx doesn't! */
      fprintf (hierFile, "<a name=\"Index%d\">&#1;</a>\n", 
				num_index_entries);
   }

   return (hi);
} /* addIndexEntry */

/* -------------------------------------------------------------------------- */

static HierIndex *newIndexEntry ()
{
HierIndex *newhi, *hi;

   newhi = (HierIndex *) RTFAlloc (sizeof (HierIndex));
#ifdef DEBUGM
   fprintf (stderr, "RTFAlloc in newIndexEntry returns: %p\n", newhi);
#endif
   if (!newhi)
      RTFPanic("Not enough memory in newIndexEntry!\n");

   newhi->next = NULL;
   newhi->heading = NULL;
   newhi->number = newhi->filenum = 0;
   strcpy(newhi->text, "");

   hi = hierIndexList;
   if (hi == NULL)
      hierIndexList = newhi;
   else {
      while (hi->next)
	 hi = hi->next;
      hi->next = newhi;
   }

   return (newhi);
} /* newIndexEntry */

/* -------------------------------------------------------------------------- */

void generateIndex ()
/* creates a file for the index */
{
FILE *fd;
char fname[256], fc = 0, ofc = 0;
int i, ofn;
HierIndex *hi;
HierHeading *hh;

   sprintf (fname, "%s-Index.html", FPrefix);

   fd = fopen(fname, "w");
   if (fd == NULL) {
      RTFPanic ("Open of %s Failed", fname);
   }

#ifdef DEBUG
   fprintf (stderr, "Generating Index...\n");
#endif

   fprintf (fd, "<html><head>\n");
   fprintf (fd, "%s", hierCreationString ());
   fprintf (fd, "<title> %s - %s </title></head><body>\n", docTitle, navText[NAV_INDEX]);

   ofn = file_number;
   file_number = -2;
   if (hierNavPanel (fd))
      fprintf (fd, "<hr>\n");
   fprintf (fd, "<h1> %s - %s </h1>\n", docTitle, navText[NAV_INDEX]);

   hi = sortIndex ();

   /* write index-panel */
   fprintf (fd, "<h2>\n");
   while (hi) {
      fc = hi->text[0];
      if (islower (fc))
	 fc -= 0x20;
      if (fc != ofc) {
	 if (ofc)
	    fprintf (fd, " , ");
	 ofc = fc;
	 fprintf (fd, "<a href=\"#IA-%c\">%c</a>\n", fc, fc);
      }

      hi = hi->next;
   };
   fprintf (fd, "</h2><hr>\n");

   /* now write index */
   fc = ofc = 0;
   hi = hierIndexList;
   fprintf (fd, "<dl><ul>\n");
   for (i = 0; i < num_index_entries; i++) {
      fc = hi->text[0];
      if (islower (fc))
	 fc -= 0x20;
      if (fc != ofc) {
	 ofc = fc;
	 fprintf (fd, "\n</ul>\n<dt><h2><a name=\"IA-%c\">%c</a></h2><ul>\n",
			fc, fc);
      }

      if (!hi->heading)
         fprintf (fd, "<li><a href=\"%s%s.html#Index%d\"> %s </a>\n", 
		FPrefix, fsuffix (hi->filenum), hi->number, hi->text);
      else {
         hh = hi->heading;
         if (hh->level <= hierDepth)
            fprintf (fd, "<li><a href=\"%s%s.html\"> %s </a>\n", 
			FPrefix, fsuffix (hh->filenum), hi->text);
         else
            fprintf (fd, "<li><a href=\"%s%s.html#Heading%d\"> %s </a>\n", 
			FPrefix, fsuffix (hh->filenum), hh->number, hi->text);
      }
      hi = hi->next;
   }

   fprintf (fd, "</ul></dl><hr>\n");   
   hierNavPanel (fd);
   file_number = ofn;

   fprintf (fd, "</body></html>\n");   
   fclose (fd);

} /*generateIndex */

/* -------------------------------------------------------------------------- */

static HierIndex *sortIndex ()
/* There certainly are smarter sorting algorithms, but I really don't care
   too much about that at the moment... */
{
HierIndex *sortedList, *hi, *hj, *hp;

   sortedList = NULL;

   while (hierIndexList) {
      hi = hierIndexList;
      hierIndexList = hi->next;
      hi->next = NULL;

      if (!sortedList)
	 sortedList = hi;
      else {
	 hj = sortedList;
	 hp = NULL;
	 while (hj) {
	    if (alpha_strcmp (hi->text, hj->text) < 0) {
	       if (!hp)
		  sortedList = hi;
	       else
	          hp->next = hi;
	       hi->next = hj;
	       break;
	    }
	    hp = hj;
	    hj = hj->next;
	 };
	 if (!hj)
	    hp->next = hi;
      }
   };

   return ((hierIndexList = sortedList));
} /* sortIndex */

/* -------------------------------------------------------------------------- */

static int alpha_strcmp (s1, s2)
   char *s1, *s2;
/* comparing function for sortIndex */
{
char c1, c2;

   while (*s1 && *s2) {
      if (*s1 != *s2) {
	 c1 = *s1;
	 c2 = *s2;

	 if (islower (c1))
	    c1 -= 0x20;
	 if (islower (c2))
	    c2 -= 0x20;

         if (c1 < c2)
	    return (-1);
	 if (c1 > c2)
	    return (1);
      }

      s1++;
      s2++;
   };

   if (!*s1 && !*s2)
      return (0);
   if (!*s1)
      return (-1);
   else
      return (1);
} /* alpha_strcmp */

/* -------------------------------------------------------------------------- */

#endif /* CB_ADD */
